import asyncio
import logging
import time
from concurrent.futures import Future
from threading import Thread
from typing import Iterable, Union, Optional

from confluent_kafka import Producer

from eventy.messaging import RecordProducer, AioRecordProducer, ProducerProduceError, RecordRouter
from eventy.record import Record
from eventy.serialization import RecordSerializer

logger = logging.getLogger(__name__)


class ConfluentKafkaRecordProducer(RecordProducer):
    """
    Confluent-Kafka producer wrapper.
    """

    def __init__(
        self,
        serializer: RecordSerializer,
        router: RecordRouter,
        bootstrap_servers: Union[str, Iterable[str]],
        sasl_username: Optional[str] = None,
        sasl_password: Optional[str] = None,
        **extra_producer_config,
    ):
        self._router = router
        self._serializer = serializer
        if isinstance(bootstrap_servers, Iterable):
            bootstrap_servers = ','.join(bootstrap_servers)
        producer_config = {
            'bootstrap.servers': bootstrap_servers,
            'enable.idempotence': True,
        }
        if sasl_username and sasl_password:
            producer_config.update(
                {
                    'sasl_mechanism': 'PLAIN',
                    'sasl_plain_username': sasl_username,
                    'sasl_plain_password': sasl_password,
                }
            )
        if extra_producer_config:
            producer_config.update(extra_producer_config)
        self._confluent_producer = Producer(producer_config)
        self._cancelled = False
        self._poll_thread = Thread(target=self._poll_loop)
        self._poll_thread.start()

    def _poll_loop(self):
        while not self._cancelled:
            self._confluent_producer.poll(0.1)

    def close(self):
        self._cancelled = True
        self._poll_thread.join()

    def produce(self, record: Record):
        # message to send
        key = record.partition_key
        value = self._serializer.encode(record)
        topic = self._router.route(record)
        logger.debug(f"Will send {record.qualified_name} on topic {topic} with key {key}.")

        # handle callback
        future_result: Future = Future()

        def ack(err, msg):
            if err:
                future_result.set_exception(ProducerProduceError(err))
            else:
                future_result.set_result(msg)

        # produce message, will trigger callback
        self._confluent_producer.produce(topic=topic, value=value, key=key, on_delivery=ack)

        # wait callback
        while not future_result.done():
            time.sleep(0.05)

        # handle exceptions
        if exception := future_result.exception():
            raise exception


class AioConfluentKafkaRecordProducer(AioRecordProducer):
    """
    Confluent-Kafka producer wrapper.
    """

    def __init__(
        self,
        serializer: RecordSerializer,
        router: RecordRouter,
        bootstrap_servers: Union[str, Iterable[str]],
        sasl_username: Optional[str] = None,
        sasl_password: Optional[str] = None,
        **extra_producer_config,
    ):
        self._router = router
        self._serializer = serializer
        if isinstance(bootstrap_servers, Iterable):
            bootstrap_servers = ','.join(bootstrap_servers)
        producer_config = {
            'bootstrap.servers': bootstrap_servers,
            'enable.idempotence': True,
        }
        if sasl_username and sasl_password:
            producer_config.update(
                {
                    'sasl_mechanism': 'PLAIN',
                    'sasl_plain_username': sasl_username,
                    'sasl_plain_password': sasl_password,
                }
            )
        if extra_producer_config:
            producer_config.update(extra_producer_config)
        self._confluent_producer = Producer(producer_config)
        self._cancelled = False

    async def start(self) -> None:
        self._loop = asyncio.get_event_loop()
        self._poll_thread = Thread(target=self._poll_loop)
        self._poll_thread.start()

    def _poll_loop(self):
        while not self._cancelled:
            self._confluent_producer.poll(0.1)

    async def close(self):
        self._cancelled = True
        self._poll_thread.join()

    async def produce(self, record: Record):
        # message to send
        key = record.partition_key
        value = self._serializer.encode(record)
        topic = self._router.route(record)
        logger.debug(f"Will send {record.qualified_name} on topic {topic} with key {key}.")

        # asyncio future
        future_result = self._loop.create_future()

        def ack(err, msg):
            logger.warning(f"Callback {err} {msg}")
            if err:
                self._loop.call_soon_threadsafe(
                    future_result.set_exception, ProducerProduceError(err)
                )
            else:
                self._loop.call_soon_threadsafe(
                    future_result.set_result, msg
                )

        self._confluent_producer.produce(topic=topic, value=value, key=key, on_delivery=ack)

        await future_result



# Copyright (c) Qotto, 2021

"""
Implementation of the eventy messaging API using Confluent Python Client, which is a wrapper around librdkafka
"""

from eventy.messaging.confluent_kafka.producer import RecordProducer, AioRecordProducer

__all__ = [
    'RecordProducer',
    'AioRecordProducer',
]



from eventy.messaging.base import RecordProducer
from eventy.record import Record
from eventy.serialization import RecordSerializer


class MemoryRecordProducer(RecordProducer):
    def __init__(
        self,
        serializer: RecordSerializer,
    ):
        self._serializer = serializer
        self._serialized_list: list[bytes] = []

    def produce(self, record: Record) -> None:
        self._serialized_list.append(self._serializer.encode(record))

    def get_produced_records(self) -> list[Record]:
        return list(map(self._serializer.decode, self._serialized_list))

    def pop_produced_records(self) -> list[Record]:
        result = self.get_produced_records()
        self._serialized_list = []
        return result



from typing import Optional

from eventy.record import Record


class RecordProducer:

    def start(self) -> None:
        pass

    def close(self) -> None:
        pass

    def produce(self, record: Record) -> None:
        raise NotImplementedError


class AioRecordProducer:

    async def start(self) -> None:
        pass

    async def close(self) -> None:
        pass

    async def produce(self, record: Record) -> None:
        raise NotImplementedError


class RecordConsumer:

    def start(self) -> None:
        pass

    def close(self) -> None:
        pass

    def poll(self) -> Optional[Record]:
        raise NotImplementedError

    def commit(self) -> None:
        raise NotImplementedError


class AioRecordConsumer:
    async def start(self) -> None:
        pass

    async def close(self) -> None:
        pass

    async def poll(self) -> Optional[Record]:
        raise NotImplementedError

    async def commit(self) -> None:
        raise NotImplementedError


class Transaction:
    def get_consumed(self) -> Record:
        raise NotImplementedError

    def send_in_transaction(self, record: Record) -> None:
        raise NotImplementedError

    def commit(self) -> None:
        raise NotImplementedError


class RecordProcessor:
    def start(self) -> None:
        pass

    def close(self) -> None:
        pass

    def transaction(self) -> Optional[Transaction]:
        raise NotImplementedError


class AioRecordProcessor:
    async def start(self) -> None:
        pass

    async def close(self) -> None:
        pass

    async def transaction(self) -> Optional[Transaction]:
        raise NotImplementedError



from eventy.record import Record


class RecordRouter:
    """
    Interface to route record to a topic
    """

    def route(self, record: Record) -> str:
        raise NotImplementedError


class SimpleRouter(RecordRouter):
    """
    Routing all records to the same topic
    """

    def __init__(
        self,
        topic: str,
    ):
        self._topic = topic

    def route(self, record: Record) -> str:
        return self._topic


class ServiceTopicRecordRouter(RecordRouter):
    """
    Routing records according to the <service>-<suffix> eventy convention
    """

    def route(self, record: Record) -> str:
        service = record.namespace.split(':')[-1]
        suffix = {
            'EVENT': 'events',
            'REQUEST': 'requests',
            'RESPONSE': 'responses',
        }.get(record.type, 'unknown')
        return f'{service}-{suffix}'



# Copyright (c) Qotto, 2021

"""
Eventy messaging API

The messaging API itself is backend-agnostic, there is currently only a Kafka backend implemented.
"""

from eventy.messaging.base import (
    RecordProducer, AioRecordProducer,
    RecordConsumer, AioRecordConsumer,
    RecordProcessor, AioRecordProcessor,
)
from eventy.messaging.errors import (
    MessagingError,
    ProducerCreationError,
    ProducerProduceError,
    ConsumerCreationError,
    ConsumerPollError,
    ConsumerCommitError,
    ProcessorTransactionError,
)
from eventy.messaging.router import (
    RecordRouter,
    ServiceTopicRecordRouter,
)

__all__ = [
    # base
    'RecordProducer',
    'AioRecordProducer',
    'RecordConsumer',
    'AioRecordConsumer',
    'RecordProcessor',
    'AioRecordProcessor',
    # router
    'RecordRouter',
    'ServiceTopicRecordRouter',
    # errors
    'MessagingError',
    'ProducerCreationError',
    'ProducerProduceError',
    'ConsumerCreationError',
    'ConsumerPollError',
    'ConsumerCommitError',
    'ProcessorTransactionError',
]
